// Copyright (C) 1999-2008
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include <stdlib.h>
#include <math.h>
#include <float.h>

#include <unistd.h>

#include <fstream>
#include "fdstream.hpp"

#include "tcl.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "base.h"
#include "fitsimage.h"
#include "util.h"
#include "mmap.h"

void Base::binCmd(const Vector& b, const Vector& v,
		       const char* x, const char* y, const char* filter)
{
  binFactor_ = b;
  binDepth_ = 1;

  if (*channelFits) {
    (*channelFits)->setBinToFactor(b);
    (*channelFits)->setBinDepth(1);
    (*channelFits)->setBinX(x);
    (*channelFits)->setBinY(y);
    (*channelFits)->setBinFilter(filter);
    updateBin((*channelFits)->updateHist(v));
  }
}

void Base::binCmd(const Vector& b, const char* x, const char* y, 
		       const char* filter)
{
  binFactor_ = b;
  binDepth_ = 1;

  if (*channelFits) {
    (*channelFits)->setBinToFactor(b);
    (*channelFits)->setBinDepth(1);
    (*channelFits)->setBinX(x);
    (*channelFits)->setBinY(y);
    (*channelFits)->setBinFilter(filter);
    updateBin((*channelFits)->updateHistCenter());
  }
}

void Base::binCmd(const Vector& b, int d, const Vector& lim, 
		       const Vector& v,
		       const char* x, const char* y, const char* z,
		       const char* filter)
{
  binFactor_ = b;
  binDepth_ = d;

  if (*channelFits) {
    (*channelFits)->setBinToFactor(b);
    (*channelFits)->setBinDepth(d);
    (*channelFits)->setBinX(x);
    (*channelFits)->setBinY(y);
    (*channelFits)->setBinZ(z);
    (*channelFits)->setBinFilter(filter);
    (*channelFits)->setBinColMinMax(z,lim);
    updateBin((*channelFits)->updateHist(v));
  }
}

void Base::binCmd(const Vector& b, int d, const Vector& lim,
		       const char* x, const char* y, const char* z,
		       const char* filter)
{
  binFactor_ = b;
  binDepth_ = d;

  if (*channelFits) {
    (*channelFits)->setBinToFactor(b);
    (*channelFits)->setBinDepth(d);
    (*channelFits)->setBinX(x);
    (*channelFits)->setBinY(y);
    (*channelFits)->setBinZ(z);
    (*channelFits)->setBinFilter(filter);
    (*channelFits)->setBinColMinMax(z,lim);
    updateBin((*channelFits)->updateHistCenter());
  }
}

void Base::binAboutCmd(const Vector& v)
{
  if (*channelFits)
    updateBin((*channelFits)->updateHist(v));
}

void Base::binColsCmd(const char* x, const char* y, const char* z)
{
  if (*channelFits) {
    (*channelFits)->setBinX(x);
    (*channelFits)->setBinY(y);
    (*channelFits)->setBinZ(z);
    updateBin(currentFits->updateHistCursor());
  }
}

void Base::binDepthCmd(int d)
{
  binDepth_ = d;
  if (*channelFits) {
    (*channelFits)->setBinDepth(d);
    updateBin(currentFits->updateHistCursor());
  }
}

void Base::binFactorCmd(const Vector& b)
{
  Vector bb = b;
  binFactor_[0] *= bb[0];
  binFactor_[1] *= bb[1];
  if (*channelFits) {
    (*channelFits)->setBinFactor(b);
    updateBin(currentFits->updateHistCursor());
  }
}

void Base::binFactorAboutCmd(const Vector& b, const Vector& v)
{
  Vector bb = b;
  binFactor_[0] *= bb[0];
  binFactor_[1] *= bb[1];
  if (*channelFits) {
    (*channelFits)->setBinFactor(b);
    updateBin((*channelFits)->updateHist(v));
  }
}

void Base::binFactorToCmd(const Vector& b)
{
  Vector bb = b;
  binFactor_[0] = bb[0];
  binFactor_[1] = bb[1];
  if (*channelFits) {
    (*channelFits)->setBinToFactor(b);
    updateBin(currentFits->updateHistCursor());
  }
}

void Base::binFactorToAboutCmd(const Vector& b, const Vector& v)
{
  Vector bb = b;
  binFactor_[0] = bb[0];
  binFactor_[1] = bb[1];
  if (*channelFits) {
    (*channelFits)->setBinToFactor(b);
    updateBin((*channelFits)->updateHist(v));
  }
}

void Base::binFunctionCmd(FitsHist::Function func)
{
  binFunction_ = func;
  if (*channelFits) {
    (*channelFits)->setBinFunction(func);
    updateBin(currentFits->updateHistCursor());
  }
}

void Base::binBufferSizeCmd(int size)
{
  binBufferSize_ = size;
  if (*channelFits) {
    (*channelFits)->setBinBufferSize(size);
    updateBin(currentFits->updateHistCursor());
  }
}

void Base::binFilterCmd(const char* filter)
{
  if (*channelFits) {
    (*channelFits)->setBinFilter(filter);
    updateBin(currentFits->updateHistCursor());
  }
}

void Base::binToFitCmd()
{
  if (*channelFits) {
    double bf = 1/calcZoom(Vector(options->width,options->height),
			   (*channelFits)->getHistDim());

    // round up to next power of 2
    if (bf < 1) {
      (*channelFits)->setBinToFactor(Vector(1,1));
      binFactor_ = Vector(1,1);
    }
    else {
      int p=1;
      while (p<bf)
	p*=2;
      (*channelFits)->setBinToFactor(Vector(p,p));
      binFactor_ = Vector(p,p);
    }
  
    updateBin(currentFits->updateHistCursor());
  }
}

void Base::bgColorCmd(const char* color)
{
  if (bgColorName)
    delete [] bgColorName;
  bgColorName = dupstr(color);
  bgColor = getXColor(bgColorName);
  update(BASE);
}

void Base::centerCmd()
{
  centerImage();
  update(MATRIX);
}

void Base::clearCmd()
{
  unloadAllFits();
  reset();
  update(MATRIX);
}

void Base::clipMinMaxModeCmd(FrScale::MinMaxMode m)
{
  if (currentScale->mmMode() != m) {
    currentScale->setMMMode(m);
    resetClip(1);
    update(BASE);
  }
}

void Base::clipMinMaxSampleCmd(int i)
{
  if (currentScale->mmIncr() != i) {
    currentScale->setMMIncr(i);
    resetClip(1);
    update(BASE);
  }
}

void Base::clipModeCmd(float per)
{
  if (per == 100) {
    if (currentScale->clipMode() != FrScale::MINMAX) {
      currentScale->setClipMode(FrScale::MINMAX);
      resetClip(1);
      update(BASE);
    }
  }
  else {
    if (currentScale->clipMode() != FrScale::AUTOCUT || 
	currentScale->autoCutPer() != per) {
      currentScale->setAutoCutPer(per);
      currentScale->setClipMode(FrScale::AUTOCUT);
      resetClip(1);
      update(BASE);
    }
  }
}

void Base::clipModeCmd(FrScale::ClipMode m)
{
  if (currentScale->clipMode() != m) {
    currentScale->setClipMode(m);
    resetClip(1);
    update(BASE);
  }
}

void Base::clipPreserveCmd(int r)
{
  currentScale->setPreserve(r);
}

void Base::clipScopeCmd(FrScale::ClipScope s)
{
  if (currentScale->clipScope() != s) {
    currentScale->setClipScope(s);
    updateClip(1);
    update(BASE);
  }
}

void Base::clipUserCmd(double l, double h)
{
  if (currentScale->uLow() != l || currentScale->uHigh()) {
    currentScale->setULow(l);
    currentScale->setUHigh(h);
    resetClip(1);
    update(BASE);
  }
}

void Base::clipUserLowCmd(double l)
{
  if (currentScale->uLow() != l) {
    currentScale->setULow(l);
    resetClip(1);
    update(BASE);
  }
}

void Base::clipUserHighCmd(double h)
{
  if (currentScale->uHigh() != h) {
    currentScale->setUHigh(h);
    resetClip(1);
    update(BASE);
  }
}

void Base::clipZScaleContrastCmd(float c)
{
  currentScale->setZContrast(c);
  resetClip(1);
  update(BASE);
}

void Base::clipZScaleSampleCmd(int s)
{
  currentScale->setZSample(s);
  resetClip(1);
  update(BASE);
}

void Base::clipZScaleLineCmd(int l)
{
  currentScale->setZLine(l);
  resetClip(1);
  update(BASE);
}

void Base::colorScaleCmd(FrScale::ColorScaleType s)
{
  setColorScale(s);
  update(BASE);
}

void Base::DATASECCmd(int which)
{
  switch (currentScale->scanMode()) {
  case FrScale::NODATASEC:
  case FrScale::DATASEC:
  case FrScale::RESET:
    currentScale->setScanMode(which ? FrScale::DATASEC : FrScale::NODATASEC);
    break;
  case FrScale::UNODATASEC:
  case FrScale::UDATASEC:
    currentScale->setScanMode(which ? FrScale::UDATASEC : FrScale::UNODATASEC);
    break;
  }

  resetClip(1);

  update(MATRIX);
}

void Base::fitsyHasExtCmd(const char* fn)
{
  // verify that we have an ext specified
  if (fn && (fn[strlen(fn)-1] != ']')) {
    Tcl_AppendResult(interp, "0", NULL);
    return;
  }

  FitsFile* ext = new FitsFitsMMap(fn, FitsFile::EXACT);
  if (ext->isValid())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);

  delete ext;
  return;
}

void Base::getBinCursorCmd()
{
  if (*channelFits)
    printVector((*channelFits)->getHistCursor(), DEFAULT);
}

void Base::getBinColsCmd()
{
  if (*channelFits && (*channelFits)->isHist())
    if ((*channelFits)->binDepth()>1)
      Tcl_AppendResult(interp, (*channelFits)->getHistX(), " ", 
		       (*channelFits)->getHistY(), " ",
		       (*channelFits)->getHistZ(), NULL);
    else
      Tcl_AppendResult(interp, (*channelFits)->getHistX(), " ", 
		       (*channelFits)->getHistY(), NULL);

  else
    Tcl_AppendResult(interp, "", NULL);
}

void Base::getBinColsMinMaxCmd(const char* col)
{
  if (*channelFits)
    printVector((*channelFits)->getHistColMinMax(col), DEFAULT);
}

void Base::getBinColsDimCmd(const char* col)
{
  if (*channelFits)
    printVector((*channelFits)->getHistColDim(col), DEFAULT);
}

void Base::getBinDepthCmd()
{
  if (*channelFits)
    binDepth_ = (*channelFits)->binDepth();
 
  printDouble(binDepth_, DEFAULT);
}

void Base::getBinFactorCmd()
{
  if (*channelFits)
    binFactor_ = (*channelFits)->binFactor();

  printVector(binFactor_, DEFAULT);
}

void Base::getBinFunctionCmd()
{
  if (*channelFits)
    binFunction_ = (*channelFits)->binFunction();

  switch (binFunction_) {
  case FitsHist::AVERAGE:
    Tcl_AppendResult(interp, "average", NULL);
    return;
  case FitsHist::SUM:
    Tcl_AppendResult(interp, "sum", NULL);
    return;
  }
}

void Base::getBinBufferSizeCmd()
{
  if (*channelFits)
    binBufferSize_ = (*channelFits)->binBufferSize();

  printInteger(binBufferSize_);
}

void Base::getBinFilterCmd()
{
  if (*channelFits && (*channelFits)->isHist())
    Tcl_AppendResult(interp, (*channelFits)->getHistFilter(), NULL);
  else
    Tcl_AppendResult(interp, "", NULL);
}

void Base::getBinListCmd()
{
  if (*channelFits && (*channelFits)->isHist()) {
    char* cols = (*channelFits)->getHistList();
    Tcl_AppendResult(interp, cols, NULL);
    delete cols;
  }
  else
    Tcl_AppendResult(interp, "", NULL);
}

void Base::getBitpixCmd()
{
  if (currentFits)
    printInteger(currentFits->bitpix());
}

void Base::getBgColorCmd()
{
  Tcl_AppendResult(interp, bgColorName, NULL);
}

void Base::getClipCmd()
{
  printVector(Vector(currentScale->low(), currentScale->high()), DEFAULT);
}

void Base::getClipCmd(float per)
{
  FrScale::ClipMode cm = (per == 100) ? FrScale::MINMAX : FrScale::AUTOCUT;
  float ac = per;

  ostringstream str;
  str << getClip(cm, ac) << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void Base::getClipCmd(FrScale::ClipMode cm)
{
  ostringstream str;
  str << getClip(cm, currentScale->autoCutPer()) << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void Base::getClipMinMaxModeCmd()
{
  switch (currentScale->mmMode()) {
  case FrScale::AUTOSCAN:
    Tcl_AppendResult(interp, "auto", NULL);
    return;
  case FrScale::SCAN:
    Tcl_AppendResult(interp, "scan", NULL);
    return;
  case FrScale::SAMPLE:
    Tcl_AppendResult(interp, "sample", NULL);
    return;
  case FrScale::DATAMIN:
    Tcl_AppendResult(interp, "datamin", NULL);
    return;
  case FrScale::IRAFMIN:
    Tcl_AppendResult(interp, "irafmin", NULL);
    return;
  }
}

void Base::getClipMinMaxSampleCmd()
{
  printInteger(currentScale->mmIncr());
}

void Base::getClipModeCmd()
{
  switch (currentScale->clipMode()) {
  case FrScale::MINMAX:
    Tcl_AppendResult(interp, "minmax", NULL);
    break;
  case FrScale::ZSCALE:
    Tcl_AppendResult(interp, "zscale", NULL);
    break;
  case FrScale::ZMAX:
    Tcl_AppendResult(interp, "zmax", NULL);
    break;
  case FrScale::AUTOCUT:
    printDouble(currentScale->autoCutPer(), DEFAULT);
    break;
  case FrScale::USERCLIP:
    Tcl_AppendResult(interp, "user", NULL);
    break;
  }
}

void Base::getClipPreserveCmd()
{
  if (currentScale->preserve())
      Tcl_AppendResult(interp, "1", NULL);
  else
      Tcl_AppendResult(interp, "0", NULL);
}

void Base::getClipScopeCmd()
{
  switch (currentScale->clipScope()) {
  case FrScale::GLOBAL:
    Tcl_AppendResult(interp, "global", NULL);
    break;
  case FrScale::LOCAL:
    Tcl_AppendResult(interp, "local", NULL);
    break;
  }
}

void Base::getClipUserCmd()
{
  printVector(Vector(currentScale->uLow(), currentScale->uHigh()), DEFAULT);
}

void Base::getClipZScaleContrastCmd()
{
  ostringstream str;
  str << currentScale->zContrast() << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void Base::getClipZScaleSampleCmd()
{
  ostringstream str;
  str << currentScale->zSample() << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void Base::getClipZScaleLineCmd()
{
  ostringstream str;
  str << currentScale->zLine() << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void Base::getColorScaleLevelCmd(int count)
{
  if (currentFits)
    getColorScaleLevelCmd(count, currentFits->getLowDouble(),
			  currentFits->getHighDouble(),
			  currentScale->colorScaleType());		
  else
    getColorScaleLevelCmd(count, currentScale->low(), currentScale->high(),
			  currentScale->colorScaleType());		
}

void Base::getColorScaleLevelCmd(int count, double ll, double hh,
				 FrScale::ColorScaleType scaleType)
{
  InverseScale* scale;
  switch (scaleType) {
  case FrScale::LINEARSCALE:
  case FrScale::IISSCALE:
    scale = new LinearInverseScale(count, ll, hh);
    break;
  case FrScale::LOGSCALE:
    scale = new LogInverseScale(count, ll, hh);
    break;
  case FrScale::POWSCALE:
    scale = new PowInverseScale(count, ll, hh);
    break;
  case FrScale::SQRTSCALE:
    scale = new SqrtInverseScale(count, ll, hh);
    break;
  case FrScale::SQUAREDSCALE:
    scale = new SquaredInverseScale(count, ll, hh);
    break;
  case FrScale::HISTEQUSCALE:
    calcHistEqu();
    scale = new HistEquInverseScale(count, ll, hh, currentScale->histequ(),
				    HISTEQUSIZE);
    break;
  }

  ostringstream str;
  str << *scale << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
 
  delete scale;
}

void Base::getColorScaleCmd()
{
  switch (currentScale->colorScaleType()) {
  case FrScale::LINEARSCALE:
    Tcl_AppendResult(interp, "linear", NULL);
    break;
  case FrScale::LOGSCALE:
    Tcl_AppendResult(interp, "log", NULL);
    break;
  case FrScale::POWSCALE:
    Tcl_AppendResult(interp, "pow", NULL);
    break;
  case FrScale::SQRTSCALE:
    Tcl_AppendResult(interp, "sqrt", NULL);
    break;
  case FrScale::SQUAREDSCALE:
    Tcl_AppendResult(interp, "squared", NULL);
    break;
  case FrScale::IISSCALE:
    Tcl_AppendResult(interp, "linear", NULL);
    break;
  case FrScale::HISTEQUSCALE:
    Tcl_AppendResult(interp, "histequ", NULL);
    break;
  }
}

void Base::getDATASECCmd()
{
  if (currentScale->scanMode())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::getFitsCountCmd()
{
  printInteger(*currentMosaicCount * *currentSliceCount);
}

void Base::getFitsDepthCmd()
{
  if (isCube())
    printInteger(*currentSliceCount);
  else
    printInteger(1);
}

void Base::getFitsHeaderCmd(int which)
{
  int prim = which < 0 ? 1:0;
  which = abs(which);

  FitsImage* rr = NULL;
  FitsImage* ptr = *channelFits;
  while (ptr && which) {
    FitsImage* sptr = ptr;
    while (sptr && which) {
      which--;
      if (!which) {
	rr = sptr;
	break;
      }
      sptr = sptr->nextSlice();
    }
    ptr = ptr->nextMosaic();
  }

  if (rr) {
    char* h = !prim ? rr->displayHeader() : rr->displayPrimary();
    Tcl_AppendResult(interp, h, NULL);
    delete h;
  }
  else
    result = TCL_ERROR;
}

void Base::getFitsHeaderKeywordCmd(int which, const char* key)
{
  which = abs(which);

  FitsImage* rr = NULL;
  FitsImage* ptr = *channelFits;
  while (ptr && which) {
    FitsImage* sptr = ptr;
    while (sptr && which) {
      which--;
      if (!which) {
	rr = sptr;
	break;
      }
      sptr = sptr->nextSlice();
    }
    ptr = ptr->nextMosaic();
  }

  if (rr) {
    char* value = rr->getKeyword(key);
    if (value) {
      Tcl_AppendResult(interp, value, NULL);
      delete [] value;
      return;
    }
  }
  else
    result = TCL_ERROR;
}

void Base::getFitsFileNameCmd(FileNameType type)
{
  if (currentFits) {
    switch (type) {
    case ROOTBASE:
      Tcl_AppendResult(interp, currentFits->getRootBaseFileName(), NULL);
      break;
    case FULLBASE:
      Tcl_AppendResult(interp, currentFits->getFullBaseFileName(), NULL);
      break;
    case ROOT:
      Tcl_AppendResult(interp, currentFits->getRootFileName(), NULL);
      break;
    case FULL:
      Tcl_AppendResult(interp, currentFits->getFullFileName(), NULL);
      break;
    }
  }
}

void Base::getFitsFileNameCmd(int which, FileNameType type)
{
  FitsImage* rr = NULL;
  FitsImage* ptr = *channelFits;
  while (ptr && which) {
    FitsImage* sptr = ptr;
    while (sptr && which) {
      which--;
      if (!which) {
	rr = sptr;
	break;
      }
      sptr = sptr->nextSlice();
    }
    ptr = ptr->nextMosaic();
  }

  if (rr) {
    switch (type) {
    case ROOTBASE:
      Tcl_AppendResult(interp, rr->getRootBaseFileName(), NULL);
      break;
    case FULLBASE:
      Tcl_AppendResult(interp, rr->getFullBaseFileName(), NULL);
      break;
    case ROOT:
      Tcl_AppendResult(interp, rr->getRootFileName(), NULL);
      break;
    case FULL:
      Tcl_AppendResult(interp, rr->getFullFileName(), NULL);
      break;
    }
  }
}

void Base::getFitsObjectNameCmd()
{
  if (currentFits)
    Tcl_AppendResult(interp, currentFits->getObjectName(), NULL);
}

void Base::getFitsMasterCmd()
{
  if (currentFits) {
    ostringstream str;
    str << currentFits->getFullFileName() << ' ' 
	<< currentFits->fitsFile() << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }
}

void Base::getFitsMosaicCountCmd()
{
  printInteger(*currentMosaicCount);
}

void Base::getFitsSizeCmd()
{
  if (keyFits && *keyFits)
    printVector(Vector((*keyFits)->width(),(*keyFits)->height()),DEFAULT);
  else
    printVector(Vector(), DEFAULT);
}

void Base::getFitsSliceCmd()
{
  if (*channelFits)
    printInteger(slice);
  else
    printInteger(1);
}

void Base::getHeightCmd()
{
  if (currentFits)
    printInteger(currentFits->height());
}

void Base::getHistogramCmd(char* xName, char* yName)
{
  bltHist(xName,yName);
  printDouble((currentScale->max()-currentScale->min())/256., DEFAULT);
}

void Base::getInfoCmd(char* var)
{
  if (currentFits) {
    Tcl_SetVar2(interp,var,"filename",
		(char*)currentFits->getRootBaseFileName(),0);
    Tcl_SetVar2(interp,var,"objectname",(char*)currentFits->getObjectName(),0);
    Tcl_SetVar2(interp,var,"min",(char*)currentFits->getMin(),0);
    Tcl_SetVar2(interp,var,"max",(char*)currentFits->getMax(),0);
    Tcl_SetVar2(interp,var,"low",(char*)currentFits->getLow(),0);
    Tcl_SetVar2(interp,var,"high",(char*)currentFits->getHigh(),0);
  }
  else {
    Tcl_SetVar2(interp,var,"objectname","",0);
    Tcl_SetVar2(interp,var,"filename","",0);
    Tcl_SetVar2(interp,var,"min","",0);
    Tcl_SetVar2(interp,var,"max","",0);
    Tcl_SetVar2(interp,var,"low","",0);
    Tcl_SetVar2(interp,var,"high","",0);
  }

  Tcl_SetVar2(interp,var,"value","",0);
  Tcl_SetVar2(interp,var,"xImage","",0);
  Tcl_SetVar2(interp,var,"yImage","",0);
  Tcl_SetVar2(interp,var,"xPhysical","",0);
  Tcl_SetVar2(interp,var,"yPhysical","",0);
  Tcl_SetVar2(interp,var,"xAmplifier","",0);
  Tcl_SetVar2(interp,var,"yAmplifier","",0);
  Tcl_SetVar2(interp,var,"xDetector","",0);
  Tcl_SetVar2(interp,var,"yDetector","",0);

  Tcl_SetVar2(interp,var,"xWCS,","",0);
  Tcl_SetVar2(interp,var,"yWCS,","",0);
  Tcl_SetVar2(interp,var,"sysWCS,","",0);
  Tcl_SetVar2(interp,var,"xUnitWCS,","",0);
  Tcl_SetVar2(interp,var,"yUnitWCS,","",0);
}

void Base::getInfoCmd(const Vector& v, InternalSystem ref,
		      SkyFrame sky, SkyFormat format, char* var)
{
  switch (ref) {
  case CANVAS:
    getInfo(v, ref, &FitsImage::getCanvasToData, &FitsImage::getCanvasToImage,
	    sky, format, var);
    break;
  case PANNER:
    getInfo(v, ref, &FitsImage::getPannerToData, &FitsImage::getPannerToImage,
	    sky, format, var);
    break;
  }
}

void Base::getInfoClipCmd()
{
  if (currentFits) {
    Tcl_AppendElement(interp, (char*)currentFits->getLow());
    Tcl_AppendElement(interp, (char*)currentFits->getHigh());
  }
  else {
    Tcl_AppendElement(interp, "0");
    Tcl_AppendElement(interp, "0");
  }
}

void Base::getMaskColorCmd()
{
  Tcl_AppendResult(interp, maskColorName, NULL);
}

void Base::getMaskMarkCmd()
{
  if (maskMark)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::getMaskTransparencyCmd()
{
  printDouble((1-maskAlpha)*100., DEFAULT);
}

void Base::getMinMaxCmd()
{
  printVector(Vector(currentScale->min(), currentScale->max()), DEFAULT);
}

void Base::getNANColorCmd()
{
  Tcl_AppendResult(interp, nanColorName, NULL);
}

void Base::getOrientCmd()
{
  switch (orientation) {
  case NORMAL:
    Tcl_AppendResult(interp, "none", NULL);
    return;
  case XX:
    Tcl_AppendResult(interp, "x", NULL);
    return;
  case YY:
    Tcl_AppendResult(interp, "y", NULL);
    return;
  case XY:
    Tcl_AppendResult(interp, "xy", NULL);
    return;
  }
}

void Base::getPanPreserveCmd()
{
  if (preservePan)
      Tcl_AppendResult(interp, "1", NULL);
  else
      Tcl_AppendResult(interp, "0", NULL);
}

void Base::getRotateCmd(Precision p)
{
  printDouble(radToDeg(rotation), p);
}

void Base::getSmoothFunctionCmd()
{
  if (*channelFits) 
    smoothFunction_ = (*channelFits)->smoothFunction();

  switch (smoothFunction_) {
  case FitsImage::BOXCAR:
    Tcl_AppendResult(interp, "boxcar", NULL);
    return;
  case FitsImage::TOPHAT:
    Tcl_AppendResult(interp, "tophat", NULL);
    return;
  case FitsImage::GAUSSIAN:
    Tcl_AppendResult(interp, "gaussian", NULL);
    return;
  }
}

void Base::getSmoothRadiusCmd()
{
  if (*channelFits)
    smoothRadius_ = (*channelFits)->smoothRadius();

  printInteger(smoothRadius_);
}

void Base::getWCSNameCmd(CoordSystem sys)
{
  if (currentFits && currentFits->hasWCS(sys)) {
    char* wcsname = (char*)currentFits->getWCSName(sys);
    if (wcsname) {
      Tcl_AppendResult(interp, wcsname, NULL);
      return;
    }
  }
  Tcl_AppendResult(interp, NULL);
}

void Base::getWidthCmd()
{
  if (currentFits)
    printInteger(currentFits->width());
}

void Base::getZoomCmd(Precision p)
{
  printVector(zoom_, p);
}

void Base::hasAmplifierCmd()
{
  if (hasATMV())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasContourCmd()
{
  if (hasContour())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasDATAMINCmd()
{
  if (currentFits && currentFits->hasDATAMIN())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasDATASECCmd()
{
  if (currentFits && currentFits->hasDATASEC())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasDetectorCmd()
{
  if (hasDTMV())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasFitsCmd()
{
  if (*channelFits)
      Tcl_AppendResult(interp, "1", NULL);
  else
      Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasFitsBinCmd()
{
  if (*channelFits && (*channelFits)->isHist())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasFitsCubeCmd()
{
  if (isCube())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasFitsMosaicCmd()
{
  if (isMosaic())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasGridCmd()
{
  if (grid)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasIISCmd()
{
  if (isIIS())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasIRAFMINCmd()
{
  if (currentFits && currentFits->hasIRAFMIN())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasPhysicalCmd()
{
  if (hasLTMV())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasSmoothCmd()
{
  if (*channelFits)
    doSmooth_ = (*channelFits)->doSmooth();

  if (doSmooth_)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasSystemCmd(CoordSystem sys)
{
  switch (sys) {
  case IMAGE:
    Tcl_AppendResult(interp, "1", NULL);
    return;
  case PHYSICAL:
    hasPhysicalCmd();
    return;
  case AMPLIFIER:
    hasAmplifierCmd();
    return;
  case DETECTOR:
    hasDetectorCmd();
    return;
  default:
    hasWCSCmd(sys);
    return;
  }
}

void Base::hasWCSCmd(CoordSystem sys)
{
  if (hasWCS(sys))
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasWCSEquCmd(CoordSystem sys)
{
  if (hasWCSEqu(sys))
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::hasWCSLinearCmd(CoordSystem sys)
{
  if (hasWCSLinear(sys))
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Base::highliteCmd(int which)
{
  useHighlite = which ? 1 : 0;
  update(PIXMAP);
}

void Base::loadIncrDataCmd(int which, int x0, int y0, int x1, int y1)
{
  FitsImage* ptr = *channelFits;
  if (which > 0) {
    for (int i=0; i<(which-1); i++) {
      if (ptr)
	ptr = ptr->nextMosaic();
    }
  }

  if (ptr)
    ptr->setDataParams(x0,y0,x1,y1);
}

void Base::loadIncrMinMaxCmd(int which, int x0, int y0, int x1, int y1)
{
  FitsImage* ptr = *channelFits;
  if (which > 0) {
    for (int i=0; i<(which-1); i++) {
      if (ptr)
	ptr = ptr->nextMosaic();
    }
  }

  if (ptr)
    ptr->setMinMaxParams(x0,y0,x1,y1);

  resetClip(1);
}

void Base::loadIncrEndCmd()
{
  switch (currentScale->scanMode()) {
  case FrScale::NODATASEC:
  case FrScale::UNODATASEC:
  case FrScale::RESET:
    currentScale->setScanMode(FrScale::NODATASEC);
    break;
  case FrScale::DATASEC:
  case FrScale::UDATASEC:
    currentScale->setScanMode(FrScale::DATASEC);
    break;
  }

  resetClip(1);
  update(MATRIX);
}

void Base::maskClearCmd()
{
  mask.deleteAll();
  update(BASE);
}

void Base::maskColorCmd(const char* color)
{
  if (maskColorName)
    delete [] maskColorName;

  maskColorName = dupstr(color);
}

void Base::maskTransparencyCmd(float t)
{
  maskAlpha = 1-(t/100.);
  update(BASE);
}

void Base::nanColorCmd(const char* color)
{
  if (nanColorName)
    delete [] nanColorName;
  nanColorName = dupstr(color);
  nanColor = getXColor(nanColorName);
  update(BASE);
}

void Base::orientCmd(Orientation which)
{
  orientation = which;

  switch (orientation) {
  case NORMAL:
    orientationMatrix.identity();
    break;
  case XX:
    orientationMatrix = FlipX();
    break;
  case YY:
    orientationMatrix = FlipY();
    break;
  case XY:
    orientationMatrix = FlipXY();
    break;
  }

  update(MATRIX);
}

void Base::pannerCmd(int s)
{
  usePanner = s;
  updatePanner();
}

void Base::pannerCmd(char* n, int w, int h)
{
  strcpy(pannerName,n);
  pannerWidth = w;
  pannerHeight = h;

  if (pannerPixmap)
    Tk_FreePixmap(display, pannerPixmap);
  pannerPixmap = 0;

  if (pannerXImage)
    XDestroyImage(pannerXImage);
  pannerXImage = NULL;

  if (pannerWidth > 0 && pannerHeight > 0) {
    if (!(pannerPixmap = Tk_GetPixmap(display, Tk_WindowId(tkwin), 
				      pannerWidth, pannerHeight, depth))) {
      cerr << "Frame Internal Error: Unable to Create Panner Pixmap" << endl;
      exit(1);
    }

    if (!(pannerXImage = XGetImage(display, pannerPixmap, 0, 0,
				   pannerWidth, pannerHeight,
				   AllPlanes, ZPixmap))){
      cerr << "Frame Internal Error: Unable to Create Panner XImage" << endl;
      exit(1);
    }
  }

  // update panner matrices
  update(MATRIX);
}

void Base::pannerCmd(CoordSystem sys, SkyFrame sky)
{
  pannerSystem = sys;
  pannerSky = sky;
  updatePanner();
}

void Base::rotateCmd(double r)
{
  rotation += r;
  update(MATRIX);
}

void Base::rotateToCmd(double r)
{
  rotation = r;
  update(MATRIX);
}

void Base::saveFitsImageFileCmd(const char* fn, int compress)
{
  if (currentFits) {
    if (!currentFits->saveFitsImageFile(fn, compress)) {
      Tcl_AppendResult(interp, "Unable to save Fits", NULL);
      result = TCL_ERROR;
    }
  }
}

void Base::saveFitsImageChannelCmd(const char* ch, int compress)
{
  if (currentFits) {
    if (!currentFits->saveFitsImageChannel(interp, ch, compress)) {
      Tcl_AppendResult(interp, "Unable to save fits to channel ", ch, NULL);
      result = TCL_ERROR;
    }
  }
}

void Base::saveFitsImageSocketCmd(int s, int compress)
{
  if (currentFits) {
    if (!currentFits->saveFitsImageSocket(s,compress)) {
      Tcl_AppendResult(interp, "Unable to save Fits", NULL);
      result = TCL_ERROR;
    }
  }
}

void Base::saveFitsTableFileCmd(const char* fn, int compress)
{
  if (currentFits) {
    if (!currentFits->saveFitsTableFile(fn, compress)) {
      Tcl_AppendResult(interp, "Unable to save Fits", NULL);
      result = TCL_ERROR;
    }
  }
}

void Base::saveFitsTableChannelCmd(const char* ch, int compress)
{
  if (currentFits) {
    if (!currentFits->saveFitsTableChannel(interp, ch, compress)) {
      Tcl_AppendResult(interp, "Unable to save fits to channel ", ch, NULL);
      result = TCL_ERROR;
    }
  }
}

void Base::saveFitsTableSocketCmd(int s, int compress)
{
  if (currentFits) {
    if (!currentFits->saveFitsTableSocket(s, compress)) {
      Tcl_AppendResult(interp, "Unable to save Fits", NULL);
      result = TCL_ERROR;
    }
  }
}

void Base::saveArrayFileCmd(const char* fn)
{
  if (currentFits) {
    if (!currentFits->saveArrayFile(fn)) {
      Tcl_AppendResult(interp, "Unable to save Array", NULL);
      result = TCL_ERROR;
    }
  }
}

void Base::saveArrayChannelCmd(const char* ch)
{
  if (currentFits) {
    if (!currentFits->saveArrayChannel(interp, ch)) {
      Tcl_AppendResult(interp, "Unable to save Array to channel ", ch, NULL);
      result = TCL_ERROR;
    }
  }
}

void Base::saveArraySocketCmd(int s)
{
  if (currentFits) {
    if (!currentFits->saveArraySocket(s)) {
      Tcl_AppendResult(interp, "Unable to save Array", NULL);
      result = TCL_ERROR;
    }
  }
}

void Base::smoothCmd(int f, int r)
{
  doSmooth_ = 1;
  smoothFunction_ = f;
  smoothRadius_ = r;

  if (*channelFits) {
    (*channelFits)->setDoSmooth(doSmooth_);
    (*channelFits)->setSmoothFunction((FitsImage::SmoothFunction)smoothFunction_);
    (*channelFits)->setSmoothRadius(smoothRadius_);
  }

  analysisFits();
  update(BASE);
}

void Base::smoothDeleteCmd()
{
  doSmooth_ = 0;

  if (*channelFits)
    (*channelFits)->setDoSmooth(doSmooth_);

  analysisFits();
  update(BASE);
}

void Base::unloadFitsCmd()
{
  unloadAllFits();
  resetValues();
  update(MATRIX);
}

void Base::updateFitsCmd(int now)
{
  if (now)
    updateNow(BASE);
  else
    update(BASE);
}

void Base::updateMagnifierCmd()
{
  updateMagnifier();
}

void Base::updateMagnifierCmd(const Vector& v)
{
  updateMagnifier(v);
}

void Base::updatePannerCmd()
{
  pannerPush();
}

void Base::warpCmd(const Vector& v)
{
  Vector a=v; // can't use a const here
  XWarpPointer(display, None, None, 0, 0, 0, 0, (int)a[0], (int)a[1]);
}

void Base::warpToCmd(const Vector& v)
{
  Vector a =v*canvasToWindow;
  XWarpPointer(display, None, Tk_WindowId(tkwin), 0,0,0,0,(int)a[0],(int)a[1]);
}

void Base::wcsAlignCmd(int which, CoordSystem sys, SkyFrame sky)
{
  wcsAlignment = which;
  wcsSystem = sys;
  wcsSky = sky;

  alignWCS();
  update(MATRIX);
}

void Base::wcsAlignCmd(int which, CoordSystem sys, SkyFrame sky,
		       FitsImage* ptr)
{
  wcsAlignment = which;
  wcsSystem = sys;
  wcsSky = sky;

  alignWCS(ptr);
  update(MATRIX);
}

void Base::wcsAppendCmd(int fd)
{
  if (!currentFits)
    return;

  boost::fdistream str(fd);
  if (!str) {
    Tcl_AppendResult(interp, " unable to read wcs infomation", NULL);
    result = TCL_ERROR;
    return;
  }

  FitsHead* r = parseWCS(str);
  currentFits->appendWCS(r);
  delete r;
}

void Base::wcsAppendCmd(const char* fn)
{
  if (!currentFits)
    return;

  ifstream str(fn);
  if (!str) {
    Tcl_AppendResult(interp, " unable to load wcs file ", fn, NULL);
    result = TCL_ERROR;
    return;
  }

  FitsHead* r = parseWCS(str);
  currentFits->appendWCS(r);
  delete r;
}

void Base::wcsResetCmd()
{
  if (currentFits)
    currentFits->resetWCS();
}

void Base::wcsReplaceCmd(int fd)
{
  if (!currentFits)
    return;

  boost::fdistream str(fd);
  if (!str) {
    Tcl_AppendResult(interp, " unable to read wcs infomation", NULL);
    result = TCL_ERROR;
    return;
  }

  FitsHead* r = parseWCS(str);
  currentFits->replaceWCS(r);
  delete r;
}

void Base::wcsReplaceCmd(const char* fn)
{
  if (!currentFits)
    return;

  ifstream str(fn);
  if (!str) {
    Tcl_AppendResult(interp, " unable to load wcs file ", fn, NULL);
    result = TCL_ERROR;
    return;
  }

  FitsHead* r = parseWCS(str);
  currentFits->replaceWCS(r);
  delete r;
}

void Base::zoomCmd(const Vector& z)
{
  Vector az = ((Vector&)z).abs();
  zoom_[0] *= az[0];
  zoom_[1] *= az[1];

  update(MATRIX);
}

void Base::zoomToCmd(const Vector& z)
{
  zoom_ = ((Vector&)z).abs();
  update(MATRIX);
}



