// Copyright (C) 1999-2008
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "frame.h"
#include "fitsimage.h"
#include "util.h"
#include "ps.h"
#include "NaN.h"
#include "contour.h"

// Frame Member Functions

Frame::Frame(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item) 
  : FrameBase(i,c,item)
{
  fits = NULL;

  mosaicCount = 0;
  sliceCount = 0;
  mosaicType = NOMOSAIC;
  mosaicSystem = WCS;
  contour = NULL;
  
  keyFits = &fits;
  channelFits = &fits;
  currentFits = NULL;
  
  currentMosaicCount = &mosaicCount;
  currentSliceCount = &sliceCount;
  currentScale = &frScale;
  currentMosaic = &mosaicType;
  currentMosaicSystem = &mosaicSystem;

  currentcontour = &contour;

  cmapID = 1;
  bias = 0.5;
  contrast = 1.0;

  colorCount = 0;
  colorScale = NULL;
  colorCells = NULL;
  indexCells = NULL;

  iisSaveRoot = 0;
  iisInteractive = 0;
}

Frame::~Frame()
{
  if (colorScale)
    delete colorScale;

  if (colorCells)
    delete [] colorCells;

  if (indexCells)
    delete [] indexCells;
}

void Frame::alignWCS()
{
  if (!wcsAlignment || !currentFits || !hasWCS(wcsSystem)) {
    wcsOrientation = NORMAL;
    wcsOrientationMatrix.identity();
    wcsRotation = 0;
    return;
  }

  calcAlignWCS(currentFits, wcsSystem, wcsSky,
	       &wcsOrientation, &wcsOrientationMatrix, &wcsRotation);
}

void Frame::alignWCS(FitsImage* ptr)
{
  if (!wcsAlignment || !ptr || !currentFits || !hasWCS(wcsSystem)) {
    wcsOrientation = NORMAL;
    wcsOrientationMatrix.identity();
    wcsRotation = 0;
    return;
  }

  calcAlignWCS(ptr, currentFits, wcsSystem, wcsSky,
	       &wcsOrientation, &wcsOrientationMatrix, &wcsRotation, &zoom_);
}

unsigned char* Frame::fillImage(int width, int height,
				double* (FitsImage::*getMatrix)())
{
  // img
  unsigned char* img = new unsigned char[width*height*3];
  {
    unsigned char* ptr = img;
    for (int jj=0; jj<height; jj++)
      for (int ii=0; ii<width; ii++) {
	*ptr++ = (unsigned char)bgColor->red;
	*ptr++ = (unsigned char)bgColor->green;
	*ptr++ = (unsigned char)bgColor->blue;
      }	
  }

  if (!currentFits)
    return img;

  // basics
  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  FitsImage* sptr = currentFits;
  int mosaic = isMosaic();

  // variable
  double* mm = (sptr->*getMatrix)();
  FitsBound* params = sptr->getDataParams(currentScale->scanMode());
  int srcw = sptr->width();

  double ll = sptr->getLowDouble();
  double hh = sptr->getHighDouble();
  double diff = hh - ll;

  // main loop
  unsigned char* dest = img;
  for (long jj=0; jj<height; jj++) {
    for (long ii=0; ii<width; ii++, dest+=3) {

      if (mosaic) {
	sptr = currentFits;

	mm = (sptr->*getMatrix)();
	params = sptr->getDataParams(currentScale->scanMode());
	srcw = sptr->width();

	ll = sptr->getLowDouble();
	hh = sptr->getHighDouble();
	diff = hh - ll;
      }

      do {
	double xx = ii*mm[0] + jj*mm[3] + mm[6];
	double yy = ii*mm[1] + jj*mm[4] + mm[7];

	if (xx>=params->xmin && xx<params->xmax && 
	    yy>=params->ymin && yy<params->ymax) {
	  double value = sptr->getValueDouble(long(yy)*srcw + long(xx));

	  if (!isnand(value)) {
	    if (value <= ll) {
	      *(dest+2) = table[0];
	      *(dest+1) = table[1];
	      *dest = table[2];
	    }
	    else if (value >= hh) {
	      *(dest+2) = table[length*3];
	      *(dest+1) = table[length*3+1];
	      *dest = table[length*3+2];
	    }
	    else {
	      int l = (int)(((value - ll)/diff * length) + .5);
	      *(dest+2) = table[l*3];
	      *(dest+1) = table[l*3+1];
	      *dest = table[l*3+2];
	    }
	  }
	  else {
	    *(dest+2) = nanColor->blue;
	    *(dest+1) = nanColor->green;
	    *dest = nanColor->red;
	  }

	  break;
	}
	else {
	  if (mosaic) {
	    sptr = sptr->nextMosaic();

	    if (sptr) {
	      mm = (sptr->*getMatrix)();
	      params = sptr->getDataParams(currentScale->scanMode());
	      srcw = sptr->width();

	      ll = sptr->getLowDouble();
	      hh = sptr->getHighDouble();
	      diff = hh - ll;
	    }
	  }
	}
      }
      while (mosaic && sptr);
    }
  }

  return img;
}

int Frame::isIIS() 
{
  return currentFits && currentFits->isIIS();
}

BBox Frame::imageBBox()
{
  // returns imageBBox in IMAGE coords
  //   and extends edge to edge

  BBox rr;
  FitsImage* ptr = fits;
  while (ptr) {
    FitsBound* params = ptr->getDataParams(frScale.scanMode());
    Matrix mm = ptr->getWCSToRef() * dataToImage;

    Vector aa =  Vector(params->xmin,params->ymin) * mm;
    if (ptr == fits)
      rr = BBox(aa,aa);
    else
      rr.bound(aa);

    rr.bound(Vector(params->xmax,params->ymin) * mm);
    rr.bound(Vector(params->xmax,params->ymax) * mm);
    rr.bound(Vector(params->xmin,params->ymax) * mm);

    ptr = ptr->nextMosaic();
  }

  return rr;
}

void Frame::ps(Filter& filter, PSColorSpace mode, int width, int height, 
	       float scale)
{
  // update psMatrix
  updatePS(currentFits, scale, width, height);

  unsigned char* img = fillImage(width, height, &FitsImage::getmPSToData);
  unsigned char* ptr = img;

  for (long jj=0; jj<height; jj++) {
    ostringstream str;

    for (long ii=0; ii<width; ii++, ptr+=3) {
      unsigned char red = ptr[0];
      unsigned char green = ptr[1];
      unsigned char blue = ptr[2];

      switch (mode) {
      case BW:
      case GRAY:
	filter << RGB2Gray(red, green, blue);
	break;
      case RGB:
	filter << red << green << blue;
	break;
      case CMYK:
	{
	  unsigned char cyan, magenta, yellow, black;
	  RGB2CMYK(red, green, blue, &cyan, &magenta, &yellow, &black);
	  filter << cyan << magenta << yellow << black;
	}
	break;
      }
      str << filter;
    }
    str << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }

  ostringstream str;
  filter.flush(str);
  psFix(str);
  Tcl_AppendResult(interp, str.str().c_str(), NULL);

  if (img)
    delete img;
}

void Frame::reset()
{
  cmapID = 1;
  bias = 0.5;
  contrast = 1.0;

  FrameBase::reset();
}

void Frame::setBinCursor(const Vector& v)
{
  if (currentFits)
    currentFits->setBinCursor(cursor);
}

void Frame::updateBin(const Matrix& mx)
{
  // Note: cursor is in USER coords, imageCenter() in IMAGE coords
  cursor = imageCenter();

  // update bin
  crosshair *= mx;

  updateMarkerCoords(&userMarkers, mx);
  updateMarkerCoords(&catalogMarkers, mx);

  // update contours
  // main contour
  if (hasContour())
    (*currentcontour)->updateCoords(mx);

  // aux contours
  {
    Contour* ptr=auxcontours.head();
    while (ptr) {
      ptr->updateCoords(mx);
      ptr=ptr->next();
    }
  }

  // delete any previous slices
  {
    FitsImage* ptr = (*channelFits)->nextSlice();
    (*channelFits)->setNextSlice(NULL);
    while (ptr) {
      FitsImage* tmp = ptr->nextSlice();
      delete ptr;
      ptr = tmp;
    }
  }

  // finish bin
  currentFits = *channelFits;

  *currentMosaicCount = 1;
  *currentSliceCount = 1;
  slice = 1;

  // bin data cube
  {
    int bd = (*channelFits)->binDepth();
    if (bd > 1) {
      FitsImage* ptr = *channelFits;
      for (int i=1; i<bd; i++) {
	FitsImage* next = NULL;
	next = new FitsImageFitsNextHist(this, *channelFits, ptr->baseFile());
	if (next && next->isValid()) {
	  ptr->setNextSlice(next);
	  ptr = next;
	  (*currentSliceCount)++;
	}
	else {
	  if (next)
	    delete next;
	  break;
	}
      }
    }
  }

  updateBinFileNames();

  resetClip(0);
  alignWCS();
  updateNow(MATRIX); // because we have changed zoom

  // update markers call backs
  // wait til matrices have been updated so that any dialogs will print
  // the correct coords

  updateMarkerCBs(&userMarkers);
  updateMarkerCBs(&catalogMarkers);
}

void Frame::updateColorCells(unsigned short* index, 
			     unsigned char* cells, int cnt)
{
  // the colorbar widget will pass us a pointer to the indexCells

  colorCount = cnt;
  if (indexCells)
    delete [] indexCells;
  indexCells = new unsigned short[cnt];
  if (!indexCells) {
    cerr << "Frame Internal Error: Unable to Alloc indexCells" << endl;
    exit(1);
  }
  memcpy(indexCells, index, cnt*sizeof(unsigned short));

  // copy the rgb vales to the colorCells array (for postscript printing)

  if (colorCells)
    delete [] colorCells;
  colorCells = new unsigned char[cnt*3];
  if (!colorCells) {
    cerr << "Frame Internal Error: Unable to Alloc colorCells" << endl;
    exit(1);
  }
  memcpy(colorCells, cells, cnt*3);
}

void Frame::updateMagnifierMatrices(Matrix& refToMagnifier)
{
  FitsImage* ptr = fits;
  while (ptr) {
    FitsImage* sptr = ptr;
    while (sptr) {
      sptr->updateMagnifierMatrices(refToMagnifier);
      sptr = sptr->nextSlice();
    }
    ptr = ptr->nextMosaic();
  }
}

void Frame::updateMatrices()
{
  // very important! do this first
  FrameBase::updateMatrices();

  // alway identity
  Matrix rgbToRef; 
  Matrix3d threedToRef3d;
  Matrix3d refToUser3d; 
  Matrix3d userToWidget3d;
  Matrix3d userToPanner3d;

  FitsImage* ptr = fits;
  while (ptr) {
    FitsImage* sptr = ptr;
    while (sptr) {
      sptr->updateMatrices(rgbToRef, threedToRef3d,
			   refToUser, refToUser3d,
			   userToWidget, userToWidget3d,
			   widgetToCanvas, 
			   userToPanner, userToPanner3d);
      sptr = sptr->nextSlice();
    }
    ptr = ptr->nextMosaic();
  }
}

void Frame::unload()
{
  unloadFits();
}

void Frame::unloadFits()
{
  if (DebugPerf)
    cerr << "Frame::unloadFits" << endl;

  // clean up from iis if needed
  if (isIIS()) {
    frScale.setClipMode(FrScale::MINMAX);
    frScale.setMMMode(FrScale::SAMPLE);
    frScale.setULow(DEFAULTLOW);
    frScale.setUHigh(DEFAULTHIGH);
    setColorScale(FrScale::LINEARSCALE);
  }

  // reset user mode
  // if (frScale.clipMode() == FrScale::USERCLIP)
  // frScale.setClipMode(FrScale::MINMAX);

  // now, delete any fits
  {
    FitsImage* ptr = fits;
    while (ptr) {
      FitsImage* sptr = ptr->nextSlice();
      while (sptr) {
	FitsImage* stmp = sptr->nextSlice();
	delete sptr;
	sptr = stmp;
      }

      FitsImage* tmp = ptr->nextMosaic();
      delete ptr;
      ptr = tmp;
    }
    fits = NULL;
  }

  // delete any contour
  if (contour)
    delete contour;
  contour = NULL;

  FrameBase::unloadFits();
}

// Commands

void Frame::getColormapCmd()
{
  ostringstream str;
  str << cmapID << ' ' << bias << ' ' << contrast << ' ' << invert << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void Frame::getRGBChannelCmd()
{
  Tcl_AppendResult(interp, "red", NULL);
}

void Frame::getRGBViewCmd()
{
  Tcl_AppendResult(interp, "1 1 1", NULL);
}

void Frame::getRGBSystemCmd()
{
  Tcl_AppendResult(interp, "image", NULL);
}

void Frame::getTypeCmd()
{
  Tcl_AppendResult(interp, "base", NULL);
}

void Frame::getWCSAlignCmd()
{
  ostringstream str;
  str << (unsigned short*)*channelFits << ends;

  Tcl_AppendResult(interp, (wcsAlignment ? "1" : "0"), " ", 
		   coordSystemStr_[wcsSystem-IMAGE], " ",
		   skyFrameStr_[wcsSky-FK4], " ",
		   str.str().c_str(), NULL);
}

void Frame::iisCmd(int width, int height)
{
  unloadAllFits();

  // force clip mode to user between 1 to 200
  frScale.setClipMode(FrScale::USERCLIP);
  frScale.setMMMode(FrScale::SCAN);
  frScale.setULow(IISMIN);
  frScale.setUHigh(IISMAX);
  setColorScale(FrScale::IISSCALE);

  FitsImage* img = new FitsImageIIS(this, width, height);

  load(ALLOC,"",img,IMG);
}

void Frame::iisEraseCmd()
{
  if (currentFits)
    ((FitsImageIIS*)currentFits)->iisErase();
}

void Frame::iisGetCmd(char* ptr, int x, int y, int dx, int dy)
{
  if (currentFits)
    memcpy(ptr, ((FitsImageIIS*)currentFits)->iisGet(x,y), dx*dy);
}

void Frame::iisSetCmd(const char* data, int x, int y, int dx, int dy)
{
  if (currentFits)
    ((FitsImageIIS*)currentFits)->iisSet(x, y, data, dx*dy);
}

void Frame::iisWCSCmd(const Matrix& mx, const Vector& z, int zt)
{
  if (currentFits)
    ((FitsImageIIS*)currentFits)->iisWCS(mx, z, zt);
}

void Frame::sliceCmd(int s)
{
  if (fits && (s>0) && (s<=sliceCount)) {
    slice = s;

    currentFits = fits;
    for (int i=1; i<s; i++)
      currentFits = currentFits->nextSlice();
  }

  // execute any update callbacks
  updateCBMarkers();

  update(BASE);
}



