// Copyright (C) 1999-2008
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "frame3d.h"
#include "fitsimage.h"
#include "util.h"
#include "ps.h"
#include "NaN.h"

Frame3d::Frame3d(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item) 
  : Frame3dBase(i,c,item)
{
  fits = NULL;

  mosaicCount = 0;
  sliceCount = 0;
  mosaicType = NOMOSAIC;
  mosaicSystem = WCS;
  
  keyFits = &fits;
  channelFits = &fits;
  currentFits = NULL;
  
  currentMosaicCount = &mosaicCount;
  currentSliceCount = &sliceCount;
  currentScale = &frScale;
  currentMosaic = &mosaicType;
  currentMosaicSystem = &mosaicSystem;

  cmapID = 1;
  bias = 0.5;
  contrast = 1.0;

  colorCount = 0;
  colorScale = NULL;
  colorCells = NULL;
  indexCells = NULL;
}

Frame3d::~Frame3d()
{
  if (colorScale)
    delete colorScale;

  if (colorCells)
    delete [] colorCells;

  if (indexCells)
    delete [] indexCells;
}

void Frame3d::alignWCS()
{
  if (!wcsAlignment || !currentFits || !hasWCS(wcsSystem)) {
    wcsOrientation = NORMAL;
    wcsOrientationMatrix.identity();
    wcsRotation = 0;
    return;
  }

  calcAlignWCS(currentFits, wcsSystem, wcsSky,
	       &wcsOrientation, &wcsOrientationMatrix, &wcsRotation);
}

void Frame3d::alignWCS(FitsImage* ptr)
{
  if (!wcsAlignment || !ptr || !currentFits || !hasWCS(wcsSystem)) {
    wcsOrientation = NORMAL;
    wcsOrientationMatrix.identity();
    wcsRotation = 0;
    return;
  }

  Vector zoom(1,1);
  calcAlignWCS(ptr, currentFits, wcsSystem, wcsSky,
	       &wcsOrientation, &wcsOrientationMatrix, &wcsRotation, &zoom);
}

Vector3d Frame3d::imageSize()
{
  // return in IMAGE coords and extends edge to edge
  Vector3d rr(100,100,100);
  FitsImage* ptr = fits;
  if (ptr) {
    FitsBound* params = ptr->getDataParams(frScale.scanMode());
    rr =  Vector3d(params->xmax-params->xmin, params->ymax-params->ymin,
		   ptr->depth());
  }
  return rr;
}

void Frame3d::reset()
{
  cmapID = 1;
  bias = 0.5;
  contrast = 1.0;

  Frame3dBase::reset();
}

void Frame3d::updateBin(const Matrix& mx)
{
  // delete any previous slices
  {
    FitsImage* ptr = (*channelFits)->nextSlice();
    (*channelFits)->setNextSlice(NULL);
    while (ptr) {
      FitsImage* tmp = ptr->nextSlice();
      delete ptr;
      ptr = tmp;
    }
  }

  // finish bin
  currentFits = *channelFits;

  *currentMosaicCount = 1;
  *currentSliceCount = 1;
  slice = 1;

  // bin data cube
  {
    int bd = (*channelFits)->binDepth();
    if (bd > 1) {
      FitsImage* ptr = *channelFits;
      for (int i=1; i<bd; i++) {
	FitsImage* next = NULL;
	next = new FitsImageFitsNextHist(this, *channelFits, ptr->baseFile());
	if (next && next->isValid()) {
	  ptr->setNextSlice(next);
	  ptr = next;
	  (*currentSliceCount)++;
	}
	else {
	  if (next)
	    delete next;
	  break;
	}
      }
    }
  }

  updateBinFileNames();

  resetClip(0);
  updateNow(MATRIX); // because we have changed zoom
}

void Frame3d::updateColorCells(unsigned short* index, 
			     unsigned char* cells, int cnt)
{
  // the colorbar widget will pass us a pointer to the indexCells

  colorCount = cnt;
  if (indexCells)
    delete [] indexCells;
  indexCells = new unsigned short[cnt];
  if (!indexCells) {
    cerr << "Frame3d Internal Error: Unable to Alloc indexCells" << endl;
    exit(1);
  }
  memcpy(indexCells, index, cnt*sizeof(unsigned short));

  // copy the rgb vales to the colorCells array (for postscript printing)

  if (colorCells)
    delete [] colorCells;
  colorCells = new unsigned char[cnt*3];
  if (!colorCells) {
    cerr << "Frame3d Internal Error: Unable to Alloc colorCells" << endl;
    exit(1);
  }
  memcpy(colorCells, cells, cnt*3);
}

void Frame3d::updateMagnifierMatrices(Matrix& refToMagnifier)
{
  FitsImage* ptr = fits;
  while (ptr) {
    FitsImage* sptr = ptr;
    while (sptr) {
      sptr->updateMagnifierMatrices(refToMagnifier);
      sptr = sptr->nextSlice();
    }
    ptr = ptr->nextMosaic();
  }
}

void Frame3d::updateMatrices()
{
  // very important! do this first
  Frame3dBase::updateMatrices();

  // alway identity
  Matrix rgbToRef; 
  Matrix userToPanner;

  FitsImage* ptr = fits;
  while (ptr) {
    FitsImage* sptr = ptr;
    while (sptr) {
      sptr->updateMatrices(rgbToRef, threedToRef3d,
			   refToUser, refToUser3d,
			   userToWidget, userToWidget3d,
			   widgetToCanvas, 
			   userToPanner, userToPanner3d);
      sptr = sptr->nextSlice();
    }
    ptr = ptr->nextMosaic();
  }
}

void Frame3d::unload()
{
  unloadFits();
}

void Frame3d::unloadFits()
{
  if (DebugPerf)
    cerr << "Frame3d::unloadFits" << endl;

  // clean up from iis if needed
  if (isIIS()) {
    frScale.setClipMode(FrScale::MINMAX);
    frScale.setMMMode(FrScale::SAMPLE);
    frScale.setULow(DEFAULTLOW);
    frScale.setUHigh(DEFAULTHIGH);
    setColorScale(FrScale::LINEARSCALE);
  }

  // now, delete any fits
  {
    FitsImage* ptr = fits;
    while (ptr) {
      FitsImage* sptr = ptr->nextSlice();
      while (sptr) {
	FitsImage* stmp = sptr->nextSlice();
	delete sptr;
	sptr = stmp;
      }

      FitsImage* tmp = ptr->nextMosaic();
      delete ptr;
      ptr = tmp;
    }
    fits = NULL;
  }

  Frame3dBase::unloadFits();
}

// Commands

void Frame3d::getColormapCmd()
{
  ostringstream str;
  str << cmapID << ' ' << bias << ' ' << contrast << ' ' << invert << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void Frame3d::getRGBChannelCmd()
{
  Tcl_AppendResult(interp, "red", NULL);
}

void Frame3d::getRGBViewCmd()
{
  Tcl_AppendResult(interp, "1 1 1", NULL);
}

void Frame3d::getRGBSystemCmd()
{
  Tcl_AppendResult(interp, "image", NULL);
}

void Frame3d::getTypeCmd()
{
  Tcl_AppendResult(interp, "3d", NULL);
}

void Frame3d::getWCSAlignCmd()
{
  ostringstream str;
  str << (unsigned short*)*channelFits << ends;

  Tcl_AppendResult(interp, (wcsAlignment ? "1" : "0"), " ", 
		   coordSystemStr_[wcsSystem-IMAGE], " ",
		   skyFrameStr_[wcsSky-FK4], " ",
		   str.str().c_str(), NULL);
}

void Frame3d::sliceCmd(int s)
{
  if (fits && (s>0) && (s<=sliceCount)) {
    slice = s;

    currentFits = fits;
    for (int i=1; i<s; i++)
      currentFits = currentFits->nextSlice();
  }

  update(BASE);
}



