// Copyright (C) 1999-2008
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "tcl.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "frame3dtruecolor.h"
#include "fitsimage.h"
#include "util.h"
#include "ps.h"
#include "NaN.h"

Frame3dTrueColor::Frame3dTrueColor(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item)
  : Frame3dBase(i,c,item), Frame3d(i,c,item), Frame3dTrue(i,c,item)
{
  colormapData = NULL;
}

Frame3dTrueColor::~Frame3dTrueColor()
{
  if (colormapData)
    delete [] colormapData;
}

void Frame3dTrueColor::buildXImage(XImage* ximage,
				   double* (FitsImage::*getMatrix)())
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  unsigned char* img = fillImage(ximage->width, ximage->height, getMatrix);
  encodeTrueColor(img, ximage);
  delete [] img;
}

unsigned char* Frame3dTrueColor::fillImage(int width, int height,
					   double* (FitsImage::*getMatrix)())
{
  depth = zbufDepth_;

  if (width==options->width) {
    zbuf = &zbufWidget_;
    mkzbuf = &mkzbufWidget_;
  }
  else if (width==pannerWidth) {
    zbuf = &zbufPanner_;
    mkzbuf = &mkzbufPanner_;
  }
  else
    return NULL;

  switch (renderMethod_) {
  case MIP:
    return fillImageMIP(width, height, depth, getMatrix);
  case AIP:
    return fillImageAIP(width, height, depth, getMatrix);
  case THRESHOLD:
    return fillImageThreshold(width, height, depth, getMatrix);
  case BLEND:
    return fillImageBlend(width, height, depth, getMatrix);
  }
}

unsigned char* Frame3dTrueColor::fillImageBlend(int width, int height, 
						int depth,
						double* 
						(FitsImage::*getMatrix)())
{
  if (!*zbuf) {
    *zbuf = new double[width*height*depth];
    memset(*zbuf, 0, width*height*depth*sizeof(double));

    *mkzbuf = new unsigned char[width*height*depth];
    memset(*mkzbuf, 0, width*height*depth);

    FitsBound* params = fits->getDataParams(currentScale->scanMode());
    int srcw = fits->width();
    int srch = fits->height();
    int srcd = fits->depth();

    double* mm = (fits->*getMatrix)();

    double* dest = *zbuf;
    unsigned char* mkdest = *mkzbuf;
    for (int jj=0; jj<height; jj++) {
      for (int ii=0; ii<width; ii++) {

	for (int kk=0; kk<depth; kk++, dest++, mkdest++) {
	  double xx = ii*mm[0] + jj*mm[4] + kk*mm[8]  + mm[12];
	  double yy = ii*mm[1] + jj*mm[5] + kk*mm[9]  + mm[13];
	  double zz = ii*mm[2] + jj*mm[6] + kk*mm[10] + mm[14];

	  if (xx>=params->xmin && xx<params->xmax && 
	      yy>=params->ymin && yy<params->ymax &&
	      zz>=0 && zz<srcd) {

	    double value = fits->getValueDouble(long(zz)*srcw*srch + 
						long(yy)*srcw + long(xx));

	    if (!isnand(value)) {
	      *dest =value;
	      *mkdest=1;
	    }
	  }
	}
      }
    }
  }

  // we go one more deep for bg
  int dd = depth+1;

  // create Z-buffer with alpha
  // stay with floating point
  float* acc = new float[width*height*dd*4];
  memset(acc, 0, width*height*dd*4*sizeof(float));

  unsigned char* mkacc = new unsigned char[width*height];
  memset(mkacc, 0, width*height);

  {
    FitsBound* params = fits->getDataParams(currentScale->scanMode());
    int srcw = fits->width();
    int srch = fits->height();
    int srcd = fits->depth();

    double* mm = (fits->*getMatrix)();

    int length = colorScale->size() - 1;
    const unsigned char* table = colorScale->psColors();

    double ll = fits->getLowDouble();
    double hh = fits->getHighDouble();
    double diff = hh - ll;

    float* dest = acc;
    unsigned char* mkdest = mkacc;
    double* src = *zbuf;
    unsigned char* mksrc = *mkzbuf;
    for (int jj=0; jj<height; jj++) {
      for (int ii=0; ii<width; ii++, mkdest++) {
	for (int kk=0; kk<depth; kk++, dest+=4, src++, mksrc++) {
	  if (*mksrc) {
	    double value = *src;

	    if ((value>=ll) && (value<=hh)) {
	      int l = (int)(((value - ll)/diff * length) + .5);
	      float alpha = (value-ll)/diff;
	      *dest = table[l*3+2]*alpha; // red
	      *(dest+1) = table[l*3+1]*alpha; // green
	      *(dest+2) = table[l*3]*alpha; // blue
	      *(dest+3) = alpha; // alpha

	      *mkdest=1;
	    }
	  }
	}

	// bg
	*dest++ = (unsigned char)bgColor->red;
	*dest++ = (unsigned char)bgColor->green;
	*dest++ = (unsigned char)bgColor->blue;
	*dest++ = 1.;
      }
    }
  }

  // bring color forward in Z-buffer
  // use front to back blend equation
  {
    float* fg = acc;
    unsigned char* mksrc = mkacc;
    for (int jj=0; jj<height; jj++) {
      for (int ii=0; ii<width; ii++, mksrc++) {
	if (*mksrc) {
	  for (int kk=0; kk<depth; kk++) {
	    float* bg = fg+4;

	    float tt = 1-*(fg+3);
	    *bg++ = *fg++ + tt* *bg; // red
	    *bg++ = *fg++ + tt* *bg; // green
	    *bg++ = *fg++ + tt* *bg; // blue
	    *bg++ = *fg++ + tt* *bg; // alpha
	  }
	  fg+=4; // skip over last layer
	}
	else
	  fg+=(depth+1)*4;
      }
    }
  }

  // convert from float to unsigned char
  // image is in last layer of acc
  unsigned char* img = new unsigned char[width*height*3];
  memset(img, 0, width*height*3);

  {
    unsigned char* dest = img;
    float* src = acc + (depth*4);
    unsigned char* mksrc = mkacc;
    for (int jj=0; jj<height; jj++) {
      for (int ii=0; ii<width; ii++, src+=dd*4, mksrc++) {
	if (*mksrc) {
	  *dest++ = *src; // red
	  *dest++ = *(src+1); // green
	  *dest++ = *(src+2); // blue
	}
	else {
	  *dest++ = (unsigned char)bgColor->red;
	  *dest++ = (unsigned char)bgColor->green;
	  *dest++ = (unsigned char)bgColor->blue;
	}
      }
    }
  }

  if (acc)
    delete acc;

  if (mkacc)
    delete mkacc;

  return img;
}

unsigned char* Frame3dTrueColor::fillImageColor(int width, int height)
{
  unsigned char* img = new unsigned char[width*height*3];
  memset(img, 0, width*height*3);

  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  double ll = fits->getLowDouble();
  double hh = fits->getHighDouble();
  double diff = hh - ll;

  register unsigned char red = (unsigned char)bgColor->red;
  register unsigned char green = (unsigned char)bgColor->green;
  register unsigned char blue = (unsigned char)bgColor->blue; 

  unsigned char* dest = img;
  double* src = *zbuf;
  unsigned char* mksrc = *mkzbuf;

  for (int jj=0; jj<height; jj++) {
    for (int ii=0; ii<width; ii++, dest+=3, src++, mksrc++) {
      *dest = red;
      *(dest+1) = green;
      *(dest+2) = blue;

      if (*mksrc) {
	double value = *src;
	if ((value>=ll) && (value<=hh)) {
	  int l = (int)(((value - ll)/diff * length) + .5);
	  *dest = table[l*3+2]; // red
	  *(dest+1) = table[l*3+1]; // green
	  *(dest+2) = table[l*3]; // blue
	}
      }
    }
  }

  return img;
}

unsigned char* Frame3dTrueColor::fillImageAIP(int width, int height, int depth,
					      double* (FitsImage::*getMatrix)())
{
  if (!*zbuf) {
    *zbuf = new double[width*height];
    memset(*zbuf, 0, width*height*sizeof(double));

    *mkzbuf = new unsigned char[width*height];
    memset(*mkzbuf, 0, width*height);

    FitsBound* params = fits->getDataParams(currentScale->scanMode());
    int srcw = fits->width();
    int srch = fits->height();
    int srcd = fits->depth();

    double* mm = (fits->*getMatrix)();

    double* dest = *zbuf;
    unsigned char* mkdest = *mkzbuf;
    for (int jj=0; jj<height; jj++) {
      for (int ii=0; ii<width; ii++, dest++, mkdest++) {
	int cnt=0;
	double acc=0;

	for (int kk=0; kk<depth; kk++) {
	  double xx = ii*mm[0] + jj*mm[4] + kk*mm[8]  + mm[12];
	  double yy = ii*mm[1] + jj*mm[5] + kk*mm[9]  + mm[13];
	  double zz = ii*mm[2] + jj*mm[6] + kk*mm[10] + mm[14];

	  if (xx>=params->xmin && xx<params->xmax && 
	      yy>=params->ymin && yy<params->ymax &&
	      zz>=0 && zz<srcd) {

	    double value = fits->getValueDouble(long(zz)*srcw*srch + 
						long(yy)*srcw + long(xx));

	    if (!isnand(value)) {
	      cnt++;
	      acc+=value;
	    }
	  }
	}

	if (cnt) {
	  *dest =acc/cnt;
	  *mkdest=1;
	}
      }
    }
  }

  return fillImageColor(width, height);
}

unsigned char* Frame3dTrueColor::fillImageMIP(int width, int height, int depth,
					      double* (FitsImage::*getMatrix)())
{
  if (!*zbuf) {
    *zbuf = new double[width*height];
    memset(*zbuf, 0, width*height*sizeof(double));

    *mkzbuf = new unsigned char[width*height];
    memset(*mkzbuf, 0, width*height);

    FitsBound* params = fits->getDataParams(currentScale->scanMode());
    int srcw = fits->width();
    int srch = fits->height();
    int srcd = fits->depth();

    double* mm = (fits->*getMatrix)();

    double* dest = *zbuf;
    unsigned char* mkdest = *mkzbuf;
    for (int jj=0; jj<height; jj++) {
      for (int ii=0; ii<width; ii++, dest++, mkdest++) {
	double max = -DBL_MAX;

	for (int kk=0; kk<depth; kk++) {
	  double xx = ii*mm[0] + jj*mm[4] + kk*mm[8]  + mm[12];
	  double yy = ii*mm[1] + jj*mm[5] + kk*mm[9]  + mm[13];
	  double zz = ii*mm[2] + jj*mm[6] + kk*mm[10] + mm[14];

	  if (xx>=params->xmin && xx<params->xmax && 
	      yy>=params->ymin && yy<params->ymax &&
	      zz>=0 && zz<srcd) {

	    double value = fits->getValueDouble(long(zz)*srcw*srch + 
						long(yy)*srcw + long(xx));

	    if (!isnand(value) && (value>max)) {
	      *dest = max = value;
	      *mkdest=1;
	    }
	  }
	}
      }
    }
  }

  return fillImageColor(width, height);
}

unsigned char* Frame3dTrueColor::fillImageThreshold(int width, int height, 
						    int depth,
						    double* 
						    (FitsImage::*getMatrix)())
{
  if (!*zbuf) {
    *zbuf = new double[width*height];
    memset(*zbuf, 0, width*height*sizeof(double));

    *mkzbuf = new unsigned char[width*height];
    memset(*mkzbuf, 0, width*height);

    FitsBound* params = fits->getDataParams(currentScale->scanMode());
    int srcw = fits->width();
    int srch = fits->height();
    int srcd = fits->depth();

    double* mm = (fits->*getMatrix)();

    double* dest = *zbuf;
    unsigned char* mkdest = *mkzbuf;
    for (int jj=0; jj<height; jj++) {
      for (int ii=0; ii<width; ii++, dest++, mkdest++) {

	for (int kk=0; kk<depth; kk++) {
	  double xx = ii*mm[0] + jj*mm[4] + kk*mm[8]  + mm[12];
	  double yy = ii*mm[1] + jj*mm[5] + kk*mm[9]  + mm[13];
	  double zz = ii*mm[2] + jj*mm[6] + kk*mm[10] + mm[14];

	  if (xx>=params->xmin && xx<params->xmax && 
	      yy>=params->ymin && yy<params->ymax &&
	      zz>=0 && zz<srcd) {

	    double value = fits->getValueDouble(long(zz)*srcw*srch + 
						long(yy)*srcw + long(xx));

	    if (!isnand(value) && (value>=threshold_)) {
	      *dest =value;
	      *mkdest=1;
	    }
	  }
	}
      }
    }
  }

  return fillImageColor(width, height);
}

// Commands

void Frame3dTrueColor::colormapCmd(int id, float b, float c, int i, 
				 unsigned short* index, unsigned char* cells, 
				 int cnt)
{
  // first check for change
  if (cmapID == id && bias == b && contrast == c && invert == i && colorCells)
    return;

  // we got a change
  cmapID = id;
  bias = b;
  contrast = c;
  invert = i;

  updateColorCells(index, cells, cnt);
  updateColorScale();

  if (mosaicCount) {
    update(BASE); // always update
    skipUpdate = 0;
  }
  else {
    if (!skipUpdate) {
      update(BASE); // update once
      skipUpdate = 1;
    }      
  }
}

